/* pvfs_rename.c copyright (c) 1996 Rob Ross and Matt Cettei,
   all rights reserved.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as published
	by the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of 
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	Contact:  Rob Ross  	rbross@parl.eng.clemson.edu
				 Matt Cettei mcettei@parl.eng.clemson.edu

 */

static char pvfs_unlink_c[] = "$Header: /projects/cvsroot/pvfs/lib/pvfs_rename.c,v 1.4 1999/10/21 02:40:26 rbross Exp $";

#include <sys/param.h>
#include <lib.h>
#include <meta.h>

extern int errno;
extern int pvfs_checks_disabled;

int pvfs_rename(const char *oldname, const char *newname)
{
	int j,i, s, index;
	mreq request;
	mack ack;
#ifdef __USE_NFS__
	char d_p[MAXPATHLEN];
	dmeta odir,ndir;
#else
	struct sockaddr *saddr;
	char *fn;
	int64_t newfs_ino, oldfs_ino;
	char bothnames[MAXPATHLEN+MAXPATHLEN+2];
	int len, bothlen;
#endif

	if (!oldname || !newname) {
		errno = EFAULT;
		return(-1);
	}

	if (pvfs_checks_disabled) return(syscall(SYS_rename, oldname, newname));

	/* check to see if file is a pvfs file */
#ifdef __USE_NFS__
	i = check_pvfs(newname);

	j = check_pvfs(oldname);
#else
	j = pvfs_detect(oldname, &fn, &saddr, &oldfs_ino);
	strncpy(bothnames, fn, MAXPATHLEN);
	i = pvfs_detect(newname, &fn, &saddr, &newfs_ino);
	len = strlen(bothnames);
	bothlen = strlen(fn) + len + 1; /* don't count final terminator */

	/* get both strings into one happy buffer */
	strncpy(&bothnames[len+1], fn, MAXPATHLEN);
#endif
	if (i==0 && j==0) {
		/* nonpvfs stuff */
		return(syscall(SYS_rename, oldname, newname));
	}
	if (i< 0 || j < 0) {
		/* oldname or newname doesn't exist */
		errno = ENOENT;
		return(-1);
	}
	if (i == 0 || j == 0 || (newfs_ino != oldfs_ino)) {
		/* across file systems, error */
		errno = EXDEV; /* not on same file system! */
		return(-1);
	}

#ifdef __USE_NFS__
	/* setup full pathnames for pvfs files */
	if (oldname[0] != '/') {
		/* get dmeta for oldname stuff */
		if (get_dmeta(oldname, &odir) < 0) {
			PERROR("pvfs_rename: get_dmeta");
			return(-1);
		}

		/* okay, absolute pathname not specified */
		strncpy(d_p,odir.rd_path,MAXPATHLEN);
		strncat(d_p,"/",MAXPATHLEN);
		strncat(d_p,odir.sd_path,MAXPATHLEN);
		strncat(d_p,"/",MAXPATHLEN);
		/* get_dmeta takes care of stripping the .. and . modifiers */
		strncat(d_p,odir.fname,MAXPATHLEN);
	}
	else
		strncpy(d_p,newname,MAXPATHLEN);

	/* add new filename path to request */
	if (newname[0] != '/') {
		/* newname dmeta stuff */
		if (get_dmeta(newname, &ndir) < 0) {
			PERROR("pvfs_rename: get_dmeta");
			return(-1);
		}

		/* absolute path not specified */
		index = strlen(d_p);
		strncat(d_p,"*",MAXPATHLEN);
		strncat(d_p,ndir.rd_path,MAXPATHLEN);
		strncat(d_p,"/",MAXPATHLEN);
		strncat(d_p,ndir.sd_path,MAXPATHLEN);
		strncat(d_p,"/",MAXPATHLEN);
		/* get_dmeta takes care of stripping the .. and . modifiers */
		strncat(d_p,ndir.fname,MAXPATHLEN);
	}
	else
		strncat(d_p,newname,MAXPATHLEN);
#endif
	
	/* Prepare request for file system  */
	request.uid = getuid();
	request.gid = getgid();
	request.type = MGR_RENAME;
	request.req.rename.oname = 0;
	request.req.rename.nname = 0;
#ifdef __USE_NFS__
	request.dsize = strlen(d_p);
	d_p[index] = '\0';
#else
	request.dsize = bothlen;
#endif
	/* Send request to mgr */	
#ifdef __USE_NFS__
	/* both old and new names better beon the same host and port number */
	if ((ndir.host != odir.host) || (ndir.port != odir.port)) {
		errno = EXDEV; /* not on same file system! */
		PERROR("pvfs_rename");
		return(-1);
	}
	if ((s = send_mreq_dmeta(&ndir, &request, d_p, &ack)) < 0) {
		PERROR("send_mreq");
		return(-1);
	}
#else
	if ((s = send_mreq_saddr(saddr, &request, bothnames, &ack)) < 0) {
		int myeno = errno;
		PERROR("send_mreq");
		errno = myeno;
		return(-1);
	}
#endif
	return(0);
}
